import { env } from '$env/dynamic/private';
import { CacheBackend } from '$lib/server/cache-backend';
import { csrf } from '$lib/server/csrf';
import { DB } from '$lib/server/drizzle';
import { runSeeds } from '$lib/server/drizzle/seeds';
import { FileBackend } from '$lib/server/file-backend';
import { JWT } from '$lib/server/jwt';
import { Logger, LogLevel } from '$lib/server/logger';
import type { ThemeModeType } from '$lib/theme-mode';
import type { Handle } from '@sveltejs/kit';
import { sequence } from '@sveltejs/kit/hooks';
import { migrate } from 'drizzle-orm/mysql2/migrator';
import { handleSession } from 'svelte-kit-cookie-session';

const { AUTO_MIGRATE, SESSION_SECRET, SESSION_SECURE, NODE_ENV } = env;

await FileBackend.init();
await CacheBackend.init();

// Set logger to debug mode
if (NODE_ENV === 'development') {
	Logger.setLogLevel(LogLevel.DEBUG);
	Logger.debug('Debug mode enabled');
}

// Initialize the database
await DB.init();

// Migrate the database when automatic migration is enabled
if (AUTO_MIGRATE === 'true') {
	Logger.info('Running automatic database migrations');
	await migrate(DB.drizzle, { migrationsFolder: './migrations' });
	Logger.info('Database migrations complete');
}

// Initialize the JWT keys
await JWT.init();

// Run database data seeders
await runSeeds();

/**
 * Set the document theme mode from the cookies
 */
const handleThemeHook = (async ({ resolve, event }) => {
	// Take the theme from the cookies
	const cookieTheme = event.cookies.get('themeMode') as ThemeModeType;
	if (cookieTheme) {
		return await resolve(event, {
			transformPageChunk: ({ html }) => html.replace('theme-base=""', `theme-base="${cookieTheme}"`)
		});
	}

	return await resolve(event);
}) satisfies Handle;

export const handle = sequence(
	csrf(['/oauth2/token', '/oauth2/introspect', '/oauth2/device_authorization', '/oauth2/par']),
	handleSession({
		secret: SESSION_SECRET,
		cookie: {
			secure: SESSION_SECURE === 'true'
		}
	}),
	handleThemeHook
);

export async function handleError({ event, error, message, status }) {
	const request = `${event.request.method} ${event.request.url} ${status} - ${message}`;
	if (status && status >= 200 && status < 500) {
		// Don't need error trace for non-server errors
		Logger.error(request);
	} else {
		Logger.error(request, error);
	}

	return { message };
}
