import { SQL, count, desc, eq, inArray, or, sql } from 'drizzle-orm';
import { DB, auditLog, user, type AuditLog, type NewAuditLog, type User } from '../drizzle';
import {
	AuditAction,
	type AuditListItem,
	type AuditSearchClause,
	type MinimalRequestEvent
} from './types';
import type { Paginated, PaginationMeta } from '$lib/types';
import { Logger } from '../logger';
import { AdminNotificationEmail, Emails } from '../email';
import { env } from '$env/dynamic/private';
import { env as publicEnv } from '$env/dynamic/public';
import { CacheBackend } from '../cache-backend';

const FLAG_EMAIL_COOLDOWN = 1 * 60 * 1000;
const FLAG_TRESHOLD_COOLDOWN = 30 * 60 * 1000;
const FLAG_EMAIL_TRESHOLD = 3;
const AUTOFLAG = [AuditAction.MALICIOUS_REQUEST, AuditAction.THROTTLE];

export class Audit {
	protected static logger = new Logger('AUDIT');
	protected static flagEmailTimeout: ReturnType<typeof setTimeout> | undefined = undefined;

	public static async insert(
		action: AuditAction,
		comment?: string,
		user?: User,
		ip?: string,
		userAgent?: string
	) {
		const flagged = AUTOFLAG.includes(action);
		const newAuditLog: NewAuditLog = {
			action,
			content: comment,
			actorId: user?.id,
			actor_ip: ip,
			actor_ua: userAgent,
			flagged: Number(flagged)
		};

		Audit.logger.info(
			`${action} ${user?.id || '-'} (${ip || '-'}) "${userAgent || '-'}" "${comment}" ${flagged}`
		);

		if (flagged) {
			Audit.auditFlagTrigger();
		}

		await DB.drizzle.insert(auditLog).values(newAuditLog);
	}

	public static async insertRequest(
		action: AuditAction,
		{ request, getClientAddress }: MinimalRequestEvent,
		user?: User,
		comment?: string
	) {
		const ip = getClientAddress();
		const userAgent = request.headers.get('user-agent') || undefined;

		return Audit.insert(action, comment, user, ip, userAgent);
	}

	public static async auditCount(search: AuditSearchClause) {
		const [{ rowCount }] = await DB.drizzle
			.select({ rowCount: count(auditLog.id).mapWith(Number) })
			.from(auditLog)
			.leftJoin(user, eq(user.id, auditLog.actorId))
			.where(Audit.getAuditWhere(search));
		return rowCount;
	}

	public static async searchAudit(
		search: AuditSearchClause & {
			limit?: number;
			offset?: number;
		}
	) {
		const limit = search.limit || 20;
		const rowCount = await Audit.auditCount(search);

		const auditSubquery = DB.drizzle
			.select({ id: auditLog.id })
			.from(auditLog)
			.leftJoin(user, eq(user.id, auditLog.actorId))
			.limit(limit)
			.offset(search.offset || 0)
			.where(Audit.getAuditWhere(search))
			.orderBy(desc(auditLog.created_at))
			.as('auditSubquery');

		const junkList = await DB.drizzle
			.select({
				audit_log: auditLog,
				user: user
			})
			.from(auditSubquery)
			.innerJoin(auditLog, eq(auditLog.id, auditSubquery.id))
			.leftJoin(user, eq(user.id, auditLog.actorId));

		const list = Audit.mapAuditRows(junkList);
		const meta: PaginationMeta = {
			rowCount,
			pageSize: limit,
			pageCount: Math.ceil(rowCount / limit)
		};

		return { list, meta } satisfies Paginated<AuditListItem>;
	}

	private static getAuditWhere(search: AuditSearchClause) {
		const selectList: SQL<unknown>[] = [];
		if (search.actions?.length) {
			const actions = Array.isArray(search.actions)
				? search.actions
				: (search.actions as string).split(',');
			selectList.push(inArray(auditLog.action, actions));
		}

		if (search.content) {
			selectList.push(sql`lower(${auditLog.content}) LIKE ${`%${search.content.toLowerCase()}%`}`);
		}

		if (search.ip) {
			selectList.push(sql`lower(${auditLog.actor_ip}) LIKE ${`%${search.ip.toLowerCase()}%`}`);
		}

		if (search.ua) {
			selectList.push(sql`lower(${auditLog.actor_ua}) LIKE ${`%${search.ua.toLowerCase()}%`}`);
		}

		if (search.user) {
			selectList.push(eq(user.uuid, search.user));
		}

		if (search.flagged) {
			selectList.push(eq(auditLog.flagged, 1));
		}

		return or(...selectList);
	}

	private static mapAuditRows(
		rows: {
			audit_log: AuditLog;
			user?: User | null;
		}[]
	) {
		return rows.reduce<AuditListItem[]>((accum, entry) => {
			let existingEntry = accum.find((e) => e.id === entry.audit_log.id);
			if (!existingEntry) {
				existingEntry = {
					id: entry.audit_log.id,
					action: entry.audit_log.action as AuditAction,
					ip: entry.audit_log.actor_ip || undefined,
					ua: entry.audit_log.actor_ua || undefined,
					content: entry.audit_log.content || undefined,
					flagged: Boolean(entry.audit_log.flagged),
					created_at: entry.audit_log.created_at
				};
				accum.push(existingEntry);
			}

			if (entry.user) {
				existingEntry.user = {
					uuid: entry.user.uuid,
					name: entry.user.display_name
				};
			}

			return accum;
		}, []);
	}

	private static async auditFlagTrigger() {
		let flagTriggerCount = (await CacheBackend.get<number>('auditFlagTriggerCount')) ?? 0;
		let flagEmailCount = (await CacheBackend.get<number>('auditFlagEmailCount')) ?? 0;
		flagTriggerCount += 1;

		await CacheBackend.set('auditFlagTriggerCount', flagTriggerCount, FLAG_TRESHOLD_COOLDOWN);

		if (flagTriggerCount < FLAG_EMAIL_TRESHOLD || Audit.flagEmailTimeout || flagEmailCount > 1) {
			return;
		}

		Logger.verbose('Flag treshold reached, sending out audit email to administrator');
		Audit.flagEmailTimeout = setTimeout(() => {
			void Audit.sendAuditEmail(flagTriggerCount);
			Audit.flagEmailTimeout = undefined;
			void CacheBackend.set('auditFlagTriggerCount', 0, FLAG_TRESHOLD_COOLDOWN);
			void CacheBackend.set('auditFlagEmailCount', flagEmailCount + 1, FLAG_TRESHOLD_COOLDOWN);
		}, FLAG_EMAIL_COOLDOWN);
	}

	/**
	 * Send an audit warning email
	 */
	private static async sendAuditEmail(count: number) {
		if (!env.EMAIL_ADMIN) {
			return;
		}

		const content = AdminNotificationEmail(count);

		try {
			await Emails.getSender().sendTemplate(
				env.EMAIL_ADMIN,
				`Audit events on ${publicEnv.PUBLIC_SITE_NAME} may require attention`,
				content
			);
		} catch (error) {
			Logger.error('Failed to send audit email:', error);
		}
	}
}
