import { env } from '$env/dynamic/private';
import {
	type JWK,
	type CryptoKey,
	importPKCS8,
	importSPKI,
	SignJWT,
	jwtVerify,
	generateKeyPair,
	exportPKCS8,
	exportSPKI,
	exportJWK
} from 'jose';
import { v4 as uuidv4 } from 'uuid';
import { FileBackend } from './file-backend';
import { DB, jwks, type JsonKey } from './drizzle';
import { CryptoUtils } from './crypto-utils';
import { Logger } from './logger';
import { lt } from 'drizzle-orm';

const { JWT_ALGORITHM, JWT_EXPIRATION, JWT_ISSUER, JWT_KEYLENGTH } = env;
const ISSUER_EXPIRY = 365 * 24 * 60 * 60 * 1000;
const ISSUER_ROTATE = ISSUER_EXPIRY / 2;

interface AvailableWebKey extends JsonKey {
	privateKey: CryptoKey;
	publicKey: CryptoKey;
	publicKeyJWK: JWK;
}

export class JWT {
	private static keys: AvailableWebKey[] = [];

	static getPublicJWKs() {
		return this.keys.map((info) => ({
			alg: JWT_ALGORITHM,
			kid: info.uuid,
			...info.publicKeyJWK,
			use: 'sig'
		}));
	}

	static async init() {
		await JWT.clearExpiredKeys();
		await JWT.loadFromStorage();
	}

	static async issue(claims: Record<string, unknown>, subject: string, audience?: string) {
		const keyInfo = JWT.getCurrentKey();
		const sign = new SignJWT(claims)
			.setProtectedHeader({ alg: JWT_ALGORITHM, kid: keyInfo.uuid })
			.setIssuedAt()
			.setSubject(subject)
			.setExpirationTime(JWT_EXPIRATION)
			.setIssuer(JWT_ISSUER);

		if (audience) {
			sign.setAudience(audience);
		}

		return sign.sign(keyInfo.privateKey);
	}

	static async verify(token: string, subject?: string, audience?: string) {
		const { payload } = await jwtVerify(
			token,
			(header) => {
				const foundKey = JWT.keys.find((item) => item.uuid === header.kid);
				if (!foundKey) {
					throw new Error('Invalid kid header value');
				}

				return foundKey.publicKey;
			},
			{
				issuer: JWT_ISSUER,
				subject,
				audience
			}
		);
		return payload;
	}

	private static async clearExpiredKeys() {
		Logger.debug(`Loading expired JWT key information`);

		const expiredQuery = lt(jwks.expires_at, new Date());
		const expireList = await DB.drizzle.select().from(jwks).where(expiredQuery);
		if (!expireList.length) {
			Logger.debug(`There are no expired JWT keys to clean up`);
			return;
		}

		for (const expired of expireList) {
			Logger.warn(`Deleting expired JWT key pair ${expired.uuid} (${expired.fingerprint})`);
			await FileBackend.deleteDirectory(['private', expired.uuid]);
		}

		// Delete expired keys
		await DB.drizzle.delete(jwks).where(expiredQuery);
	}

	private static async loadFromStorage() {
		// Fetch existing keys
		const keys = await DB.drizzle.select().from(jwks);
		Logger.verbose(`Loading ${keys.length} JWT key pairs into memory`);
		JWT.keys = await JWT.loadKeys(keys);

		// No current key or it is time to rotate
		const keyPair = JWT.keys.find(({ current }) => current === 1);
		if (!keyPair || keyPair.rotate_at.getTime() < Date.now()) {
			Logger.verbose(`Creating a new JWT key pair`);

			const newKey = await JWT.createKeyPair();
			if (keyPair) {
				keyPair.current = 0;
			}

			JWT.keys.push(newKey);
		}
	}

	private static getCurrentKey() {
		const current = JWT.keys.find(({ current }) => current === 1);
		if (!current) {
			throw new Error('No current key found');
		}
		return current;
	}

	private static async createKeyPair() {
		/**
		 * Equivalent openssl commands:
		 * Private: openssl genpkey -out jwt.private.pem -algorithm RSA -pkeyopt rsa_keygen_bits:2048
		 * Public:  openssl rsa -in jwt.private.pem -pubout -outform PEM -out jwt.public.pem
		 */
		const { privateKey, publicKey } = await generateKeyPair(JWT_ALGORITHM, {
			modulusLength: Number(JWT_KEYLENGTH) || 2048
		});
		const jwk = await exportJWK(publicKey);
		const kid = uuidv4({ random: Buffer.from(jwk.n as string).subarray(0, 16) });

		// Save to file backend
		const exportedPrivateKey = await exportPKCS8(privateKey);
		const exportedPublicKey = await exportSPKI(publicKey);
		await FileBackend.saveFile(['private', kid, 'jwt.public.pem'], Buffer.from(exportedPublicKey));
		await FileBackend.saveFile(
			['private', kid, 'jwt.private.pem'],
			Buffer.from(exportedPrivateKey)
		);

		// Save to database
		const fingerprint = CryptoUtils.fingerprintPem(exportedPublicKey);
		const entity = {
			uuid: kid,
			current: 1,
			fingerprint,
			created_at: new Date(),
			rotate_at: new Date(Date.now() + ISSUER_ROTATE),
			expires_at: new Date(Date.now() + ISSUER_EXPIRY)
		} as JsonKey;

		await DB.drizzle.update(jwks).set({ current: 0 });
		await DB.drizzle.insert(jwks).values(entity);

		// Return full pair information
		return {
			...entity,
			privateKey,
			publicKey,
			publicKeyJWK: jwk
		} as AvailableWebKey;
	}

	private static async loadKeys(keys: JsonKey[]) {
		return Promise.all(
			keys.map(async (entity) => {
				const publicKey = await FileBackend.readFile<string>(
					['private', entity.uuid, 'jwt.public.pem'],
					'utf-8'
				);
				const privateKey = await FileBackend.readFile<string>(
					['private', entity.uuid, 'jwt.private.pem'],
					'utf-8'
				);
				const importedPrivateKey = await importPKCS8(privateKey, JWT_ALGORITHM);
				const importedPublicKey = await importSPKI(publicKey, JWT_ALGORITHM);
				const jwk = await exportJWK(importedPublicKey);

				// Return full pair information
				return {
					...entity,
					privateKey: importedPrivateKey,
					publicKey: importedPublicKey,
					publicKeyJWK: jwk
				} as AvailableWebKey;
			})
		);
	}
}
