import chalk from 'chalk';
import { format as strfmt, formatWithOptions as strfmtOptions } from 'node:util';

export enum LogLevel {
	DEBUG = 0,
	VERBOSE,
	INFO,
	WARN,
	ERROR
}

export interface ILogger {
	log: (format: any, ...arg: any[]) => void;
	debug: (format: any, ...arg: any[]) => void;
	verbose: (format: any, ...arg: any[]) => void;
	info: (format: any, ...arg: any[]) => void;
	warn: (format: any, ...arg: any[]) => void;
	error: (format: any, ...arg: any[]) => void;
}

const chalkColors: Record<LogLevel, (str: string) => string> = {
	[LogLevel.DEBUG]: chalk.magenta,
	[LogLevel.VERBOSE]: chalk.blueBright,
	[LogLevel.INFO]: chalk.green,
	[LogLevel.WARN]: chalk.yellow,
	[LogLevel.ERROR]: chalk.red
};

export class Logger implements ILogger {
	protected static instance?: Logger;

	constructor(
		public context: string = '',
		public logLevel: LogLevel = LogLevel.VERBOSE
	) {}

	static getInstance() {
		if (!Logger.instance) {
			Logger.instance = new Logger();
		}
		return Logger.instance;
	}

	protected get formatted() {
		return this.logLevel === LogLevel.DEBUG;
	}

	protected colorize(level: LogLevel, leading: string, trailing: string) {
		if (!this.formatted) {
			return leading + trailing;
		}

		return chalkColors[level](leading) + trailing;
	}

	protected handle(level: LogLevel, format: any, ...arg: any[]): void {
		if (level < this.logLevel) {
			return;
		}

		const timestamp = this.getTimestamp();
		const logLevelKey = (Object.values(LogLevel) as string[])[level].toUpperCase().padStart(7);

		let formatted = this.formatted
			? strfmtOptions({ colors: true, depth: 8 }, format, ...arg)
			: strfmt(format, ...arg);

		// Keep non-debug logs on one line for neatness
		if (!this.formatted) {
			formatted = formatted.replace(/\n/g, ' ');
		}

		const outLine = this.colorize(
			level,
			`[${process.pid}] ${timestamp} ${logLevelKey}${this.context ? ` [${this.context}]` : ''} `,
			formatted
		);
		this.writeLog(LogLevel.ERROR ? 'err' : 'out', outLine);
	}

	protected getTimestamp() {
		return new Intl.DateTimeFormat('en-GB', {
			year: 'numeric',
			month: '2-digit',
			day: '2-digit',
			hour: '2-digit',
			minute: '2-digit',
			second: '2-digit'
		}).format(new Date());
	}

	protected writeLog(stream: 'out' | 'err', message: string) {
		(stream === 'out' ? process.stdout : process.stderr).write(`${message}\r\n`);
	}

	setLogLevel(newLevel: LogLevel) {
		this.logLevel = newLevel;
	}

	static setLogLevel(newLevel: LogLevel) {
		Logger.getInstance().logLevel = newLevel;
	}

	log(format: any, ...arg: any[]) {
		this.handle(LogLevel.INFO, format, ...arg);
	}

	static log(format: any, ...arg: any[]) {
		Logger.getInstance().handle(LogLevel.INFO, format, ...arg);
	}

	debug(format: any, ...arg: any[]) {
		this.handle(LogLevel.DEBUG, format, ...arg);
	}

	static debug(format: any, ...arg: any[]) {
		Logger.getInstance().handle(LogLevel.DEBUG, format, ...arg);
	}

	verbose(format: any, ...arg: any[]) {
		this.handle(LogLevel.VERBOSE, format, ...arg);
	}

	static verbose(format: any, ...arg: any[]) {
		Logger.getInstance().handle(LogLevel.VERBOSE, format, ...arg);
	}

	info(format: any, ...arg: any[]) {
		this.handle(LogLevel.INFO, format, ...arg);
	}

	static info(format: any, ...arg: any[]) {
		Logger.getInstance().handle(LogLevel.INFO, format, ...arg);
	}

	warn(format: any, ...arg: any[]) {
		this.handle(LogLevel.WARN, format, ...arg);
	}

	static warn(format: any, ...arg: any[]) {
		Logger.getInstance().handle(LogLevel.WARN, format, ...arg);
	}

	error(format: any, ...arg: any[]) {
		this.handle(LogLevel.ERROR, format, ...arg);
	}

	static error(format: any, ...arg: any[]) {
		Logger.getInstance().handle(LogLevel.ERROR, format, ...arg);
	}
}
