import { env } from '$env/dynamic/public';
import { ApiUtils } from '$lib/server/api-utils';
import { Logger } from '$lib/server/logger';
import { InvalidClient, InvalidRequest, InvalidScope, UnauthorizedClient } from '../error';
import { OAuth2Clients, OAuth2DeviceCodes, OAuth2Tokens } from '../model';
import { OAuth2Response } from '../response';

export class OAuth2DeviceAuthorizationController {
	static async postRequest({ request }: { request: Request }) {
		const body = await ApiUtils.getJsonOrFormBody(request);

		let clientId: string | null = null;
		let clientSecret: string | null = null;
		let clientAssertionType: string | null = null;
		let clientAssertion: string | null = null;

		if (body.client_secret) {
			clientId = body.client_id as string;
			clientSecret = body.client_secret as string;
			Logger.debug('Client basic credentials parsed from body parameters', clientId, clientSecret);
		} else if (body.client_assertion) {
			clientId = body.client_id as string;
			clientAssertionType = body.client_assertion_type;
			clientAssertion = body.client_assertion;
			Logger.debug(
				'Client assertion credentials parsed from body parameters',
				clientId,
				clientAssertionType,
				clientAssertion
			);
		} else {
			if (!request.headers?.has('authorization')) {
				throw new InvalidRequest('No authorization header passed');
			}

			let pieces = (request.headers.get('authorization') as string).split(' ', 2);
			if (!pieces || pieces.length !== 2) {
				throw new InvalidRequest('Authorization header is corrupted');
			}

			if (pieces[0] !== 'Basic') {
				throw new InvalidRequest(`Unsupported authorization method: ${pieces[0]}`);
			}

			pieces = Buffer.from(pieces[1], 'base64').toString('ascii').split(':', 2);
			if (!pieces || pieces.length !== 2) {
				throw new InvalidRequest('Authorization header has corrupted data');
			}

			clientId = pieces[0];
			clientSecret = pieces[1];
			Logger.debug('Client credentials parsed from basic auth header:', clientId, clientSecret);
		}

		if (!clientId) {
			throw new InvalidRequest('client_id field is mandatory for device authorization endpoint');
		}

		const client = await OAuth2Clients.fetchById(clientId);
		if (!client || client.activated === 0) {
			throw new InvalidClient('Client not found');
		}

		if (client.confidential === 1 || clientSecret || clientAssertion) {
			if (clientAssertion && clientAssertionType) {
				const valid = await OAuth2Clients.validateClientAssertionAuthentication(
					client,
					clientAssertionType,
					clientAssertion
				);
				if (!valid) {
					throw new UnauthorizedClient('Invalid client assertion');
				}
			} else {
				const valid = OAuth2Clients.checkSecret(client, clientSecret);
				if (!valid) {
					throw new UnauthorizedClient('Invalid client secret');
				}
			}
		}

		if (!OAuth2Clients.checkGrantType(client, 'device_code')) {
			throw new UnauthorizedClient('This client does not support grant type device');
		}

		const scope = OAuth2Clients.transformScope(body.scope || '');
		if (!OAuth2Clients.checkScope(client, scope)) {
			throw new InvalidScope('Client does not allow access to this scope');
		}

		const issued = await OAuth2DeviceCodes.create(clientId, scope);

		return OAuth2Response.createResponse(200, {
			...issued,
			verification_uri: `${env.PUBLIC_URL}/device`,
			verification_uri_complete: `${env.PUBLIC_URL}/device?user_code=${issued.user_code}`,
			expires_in: OAuth2Tokens.deviceTtl,
			interval: OAuth2DeviceCodes.interval
		});
	}
}
