import { ApiUtils } from '$lib/server/api-utils';
import { Logger } from '$lib/server/logger';
import {
	InvalidRequest,
	UnsupportedResponseType,
	InvalidClient,
	UnauthorizedClient,
	InvalidScope,
	InvalidGrant
} from '../error';
import { OAuth2Clients, OAuth2ParCodes, OAuth2Tokens, type CodeChallengeMethod } from '../model';
import { OAuth2Response } from '../response';

export class OAuth2PushedAuthorizationController {
	static async postRequest({ request }: { request: Request }) {
		const body = await ApiUtils.getJsonOrFormBody(request);

		if (!body.redirect_uri) {
			throw new InvalidRequest('redirect_uri field is mandatory for authorization endpoint');
		}

		const redirectUri = body.redirect_uri as string;
		Logger.debug('Parameter redirect uri is', redirectUri);

		let clientId: string | null = null;
		let clientSecret: string | null = null;
		let clientAssertionType: string | null = null;
		let clientAssertion: string | null = null;

		if (body.client_secret) {
			clientId = body.client_id as string;
			clientSecret = body.client_secret as string;
			Logger.debug('Client basic credentials parsed from body parameters', clientId, clientSecret);
		} else if (body.client_assertion) {
			clientId = body.client_id as string;
			clientAssertionType = body.client_assertion_type;
			clientAssertion = body.client_assertion;
			Logger.debug(
				'Client assertion credentials parsed from body parameters',
				clientId,
				clientAssertionType,
				clientAssertion
			);
		} else {
			if (!request.headers?.has('authorization')) {
				throw new InvalidRequest('No authorization header passed');
			}

			let pieces = (request.headers.get('authorization') as string).split(' ', 2);
			if (!pieces || pieces.length !== 2) {
				throw new InvalidRequest('Authorization header is corrupted');
			}

			if (pieces[0] !== 'Basic') {
				throw new InvalidRequest(`Unsupported authorization method: ${pieces[0]}`);
			}

			pieces = Buffer.from(pieces[1], 'base64').toString('ascii').split(':', 2);
			if (!pieces || pieces.length !== 2) {
				throw new InvalidRequest('Authorization header has corrupted data');
			}

			clientId = pieces[0];
			clientSecret = pieces[1];
			Logger.debug('Client credentials parsed from basic auth header:', clientId, clientSecret);
		}

		if (!clientId) {
			throw new InvalidRequest('client_id field is mandatory for authorization endpoint');
		}

		if (!body.response_type) {
			throw new InvalidRequest('response_type field is mandatory for authorization endpoint');
		}

		const responseType = body.response_type as string;
		Logger.debug('Parameter response_type is', responseType);

		// Support multiple types
		const responseTypes = responseType.split(' ');
		let grantTypes: string[] = [];
		for (const responseType of responseTypes) {
			switch (responseType) {
				case 'code':
					grantTypes.push('authorization_code');
					break;
				case 'token':
					grantTypes.push('implicit');
					break;
				case 'id_token':
				case 'none':
					grantTypes.push(responseType);
					break;
				default:
					throw new UnsupportedResponseType('Unknown response_type parameter passed');
			}
		}

		// Filter out duplicates
		grantTypes = grantTypes.filter((value, index, self) => self.indexOf(value) === index);

		// "None" type cannot be combined with others
		if (grantTypes.length > 1 && grantTypes.includes('none')) {
			throw new InvalidRequest('Grant type "none" cannot be combined with other grant types');
		}

		Logger.debug('Parameter grant_type is', grantTypes.join(' '));

		const client = await OAuth2Clients.fetchById(clientId);
		if (!client || client.activated === 0) {
			throw new InvalidClient('Client not found');
		}

		if (client.confidential === 0) {
			throw new InvalidRequest(
				'Non-confidential clients cannot use Pushed Authorization Requests, as the origin of such requests cannot be verified'
			);
		}

		if (clientAssertion && clientAssertionType) {
			const valid = await OAuth2Clients.validateClientAssertionAuthentication(
				client,
				clientAssertionType,
				clientAssertion
			);
			if (!valid) {
				throw new UnauthorizedClient('Invalid client assertion');
			}
		} else {
			const valid = OAuth2Clients.checkSecret(client, clientSecret);
			if (!valid) {
				throw new UnauthorizedClient('Invalid client secret');
			}
		}
		Logger.debug('client secret check passed');

		if (!(await OAuth2Clients.getRedirectUrls(client.client_id))?.length) {
			throw new UnsupportedResponseType('The client has not set a redirect uri');
		} else if (!(await OAuth2Clients.checkRedirectUri(client, redirectUri))) {
			throw new InvalidRequest('Wrong RedirectUri provided');
		}
		Logger.debug('redirect_uri check passed');

		// The client needs to support all grant types
		for (const grantType of grantTypes) {
			if (grantType !== 'none' && !OAuth2Clients.checkGrantType(client, grantType)) {
				throw new UnauthorizedClient('This client does not support grant type ' + grantType);
			}
		}
		Logger.debug('Grant type check passed');

		const scope = OAuth2Clients.transformScope(body.scope as string);
		if (!OAuth2Clients.checkScope(client, scope)) {
			throw new InvalidScope('Client does not allow access to this scope');
		}
		Logger.debug('Scope check passed');

		const codeChallenge = body.code_challenge as string;
		const codeChallengeMethod = (body.code_challenge_method as CodeChallengeMethod) || 'plain';

		if (codeChallengeMethod && !OAuth2Tokens.challengeMethods.includes(codeChallengeMethod)) {
			throw new InvalidGrant('Invalid code challenge method');
		}

		if (client.confidential === 0 && !codeChallenge && grantTypes.includes('authorization_code')) {
			throw new InvalidGrant(
				'A code_challenge is required for the authorization_code grant in non-confidential clients'
			);
		}

		const nonce = body.nonce as string;
		const state = body.state as string;

		const response = await OAuth2ParCodes.create(
			clientId,
			scope,
			nonce,
			codeChallenge,
			codeChallengeMethod,
			state,
			redirectUri,
			grantTypes.join(',')
		);

		return OAuth2Response.createResponse(200, response);
	}
}
