import type { OAuth2Client } from '$lib/server/drizzle';
import { Logger } from '$lib/server/logger';
import { Users } from '$lib/server/users';
import { AccessDenied, AuthorizationPending, ExpiredToken, ServerError } from '../../error';
import {
	OAuth2AccessTokens,
	OAuth2Clients,
	OAuth2DeviceCodes,
	OAuth2RefreshTokens,
	OAuth2Tokens,
	OAuth2Users
} from '../../model';
import type { OAuth2TokenResponse } from '../../response';

export async function device(client: OAuth2Client, deviceCode: string) {
	const token = await OAuth2DeviceCodes.getByDeviceCode(deviceCode, client.client_id);
	if (!token) {
		throw new AccessDenied('The Device Code is invalid or the request was denied');
	}

	if (!OAuth2Tokens.checkTTL(token)) {
		throw new ExpiredToken('The Device Code is expired');
	}

	if (!token.userId) {
		throw new AuthorizationPending('Authorization is pending');
	}

	const cleanScope = OAuth2Clients.transformScope(token.scope || '');

	const user = await Users.getById(token.userId);
	if (!user) {
		throw new ServerError('The user was not found');
	}

	const resObj: OAuth2TokenResponse = {
		token_type: 'bearer'
	};

	try {
		resObj.access_token = await OAuth2AccessTokens.create(
			user.id,
			client.client_id,
			cleanScope,
			OAuth2Tokens.tokenTtl
		);
	} catch {
		throw new ServerError('Failed to call accessTokens.create method');
	}

	if (cleanScope.includes('openid')) {
		try {
			resObj.id_token = await OAuth2Users.issueIdToken(user, client, cleanScope);
		} catch (err) {
			Logger.error('Failed to issue a new ID token:', err);
			throw new ServerError('Failed to issue an ID token');
		}
	}

	if (OAuth2Clients.checkGrantType(client, 'refresh_token')) {
		try {
			resObj.refresh_token = await OAuth2RefreshTokens.create(
				user.id,
				client.client_id,
				cleanScope
			);
		} catch (err) {
			throw new ServerError('Failed to call refreshTokens.create function');
		}
	}

	resObj.expires_in = OAuth2Tokens.tokenTtl;

	await OAuth2DeviceCodes.removeByCode(deviceCode);

	return resObj;
}
