import { browser } from '$app/environment';
import { onMount } from 'svelte';
import { writable } from 'svelte/store';
import { THEME_COOKIE_MAX_AGE } from './constants';

export type ThemeModeType = 'light' | 'dark';

export const themeMode = writable<ThemeModeType>('light', (set) => {
	if (!browser) return;

	set((document.documentElement.getAttribute('theme-base') as ThemeModeType) || 'light');
});

/**
 * Save theme as a cookie.
 * @param theme - theme to save
 */
export const setThemeCookie = (theme: ThemeModeType) =>
	(document.cookie = `themeMode=${theme}; path=/; SameSite=Lax; Max-Age=${THEME_COOKIE_MAX_AGE}`);

/**
 * Forward the initial theme to the theme store, and subscribe to changes.
 */
export const forwardInitialTheme = () =>
	onMount(() => {
		// If a cookie-set theme is not present, set the theme to current user agent theme.
		const hasServerTheme = !!document.documentElement.getAttribute('theme-base');
		const uaTheme = window?.matchMedia?.('(prefers-color-scheme: dark)');
		const uaMode: ThemeModeType = uaTheme?.matches ? 'dark' : 'light';

		if (!hasServerTheme) {
			themeMode.set(uaMode);
		}

		// Set the theme to the current theme store value, and set the cookie.
		const unsubscribeForward = themeMode.subscribe(async (theme) => {
			document.documentElement.setAttribute('theme-base', theme);
			setThemeCookie(theme);
		});

		// Set the theme to the user agent theme when the user agent theme changes.
		const uaSetter = () => {
			themeMode.set(uaTheme?.matches ? 'dark' : 'light');
		};

		uaTheme?.addEventListener?.('change', uaSetter);
		return () => {
			unsubscribeForward();
			uaTheme?.removeEventListener('change', uaSetter);
		};
	});
