import { OAUTH2_MAX_REDIRECTS, OAUTH2_MAX_URLS } from '$lib/constants.js';
import { AdminUtils } from '$lib/server/admin-utils';
import { Audit, AuditAction } from '$lib/server/audit';
import { Changesets } from '$lib/server/changesets.js';
import { CryptoUtils } from '$lib/server/crypto-utils.js';
import type { OAuth2Client, User } from '$lib/server/drizzle';
import {
	OAuth2ClientURLType,
	OAuth2Clients,
	type OAuth2ClientAdminListItem
} from '$lib/server/oauth2';
import { Uploads } from '$lib/server/upload.js';
import { Users } from '$lib/server/users';
import { hasPrivileges } from '$lib/utils';
import { emailRegex, privilegeRegex } from '$lib/validators.js';
import { error, fail, redirect } from '@sveltejs/kit';
import { RateLimiter } from 'sveltekit-rate-limiter/server';

interface AddUrlRequest {
	type: OAuth2ClientURLType;
	url: string;
}

interface UpdateRequest {
	title: string;
	description: string;
	activated?: string;
	verified?: string;
	confidential?: string;
}

interface AddPrivilegeRequest {
	name: string;
}

interface InviteRequest {
	email: string;
}

const oneOffLimiter = new RateLimiter({
	IP: [1, 'm']
});

const inviteLimiter = new RateLimiter({
	IP: [3, 'm']
});

/**
 * Return the current user's information and the client the action is performed on.
 * Errors if client doesn't exist or user does not have permission to change it.
 * @param locals App locals
 * @param uuid Client ID
 * @returns Current user, their session, full OAuth2 privileges flag and the actionable client
 */
const getActionData = async (locals: App.Locals, uuid: string) => {
	const { currentUser, userSession } = await AdminUtils.getActionUser(locals, [
		'admin:oauth2',
		'self:oauth2'
	]);

	const fullPrivileges = hasPrivileges(userSession.privileges || [], ['admin:oauth2']);

	const {
		list: [details]
	} = await OAuth2Clients.getClientByAdminUser(currentUser as User, {
		clientId: uuid,
		listAll: fullPrivileges,
		omitSecret: false
	});

	if (!details) {
		return error(404, 'Client not found');
	}

	return { currentUser, userSession, fullPrivileges, details: details as OAuth2Client };
};

export const actions = {
	/**
	 * Update the OAuth2 Client general information.
	 */
	update: async ({ locals, request, params: { uuid }, getClientAddress }) => {
		const { details, fullPrivileges, currentUser } = await getActionData(locals, uuid);

		const body = await request.formData();
		const { title, description, activated, verified, confidential } = Changesets.only(
			['title', 'description', 'activated', 'verified', 'confidential'],
			body
		);

		if (!!verified && !fullPrivileges) {
			return fail(403, { errors: ['forbidden'] });
		}

		const actuallyVerified = fullPrivileges ? Number(!!verified) : undefined;
		const actuallyConfidential = fullPrivileges ? Number(!!confidential) : undefined;
		const actuallyActivated = Number(!!activated);

		if (title && (title.length < 3 || title.length > 32)) {
			return fail(403, { errors: ['invalidTitle'] });
		}

		if (description && description.length > 1000) {
			return fail(403, { errors: ['invalidDescription'] });
		}

		await OAuth2Clients.update(details, {
			title,
			description,
			verified: actuallyVerified,
			activated: actuallyActivated,
			confidential: actuallyConfidential
		});

		await Audit.insertRequest(
			AuditAction.OAUTH2_UPDATE,
			{ request, getClientAddress },
			currentUser,
			`client_id=${details.client_id}`
		);

		return { errors: [] };
	},
	/**
	 * Permanently delete the OAuth2 Client.
	 *
	 * Only works for disabled clients, and only the owner or admin can do it.
	 */
	delete: async ({ locals, params: { uuid }, ...event }) => {
		const { currentUser, fullPrivileges, details } = await getActionData(locals, uuid);

		if (details.activated === 1) {
			return fail(400, { errors: ['deleteActivated'] });
		}

		if (details.ownerId !== currentUser.id && !fullPrivileges) {
			return fail(403, { errors: ['forbidden'] });
		}

		await OAuth2Clients.deleteClient(details);

		await Audit.insertRequest(
			AuditAction.OAUTH2_DELETE,
			event,
			currentUser,
			`client_id=${details.client_id}\ntitle=${details.title}`
		);

		return redirect(303, '/ssoadmin/oauth2');
	},
	/**
	 * Regenerate OAuth2 Client secret key.
	 *
	 * Only the owner or admin can do it.
	 */
	regenerate: async (event) => {
		const {
			locals,
			params: { uuid }
		} = event;
		const { currentUser, fullPrivileges, details } = await getActionData(locals, uuid);

		if (!fullPrivileges && details.ownerId !== currentUser.id) {
			return fail(403, { errors: ['forbidden'] });
		}

		// Allow secret regeneration only once per minute.
		if (!fullPrivileges && (await oneOffLimiter.isLimited(event))) {
			await Audit.insertRequest(
				AuditAction.THROTTLE,
				event,
				currentUser,
				`secret regeneration attempt`
			);
			return fail(429, { errors: ['tooManyTimes'] });
		}

		await OAuth2Clients.update(details, {
			client_secret: CryptoUtils.generateSecret()
		});

		await Audit.insertRequest(
			AuditAction.OAUTH2_REGENERATE,
			event,
			currentUser,
			`client_id=${details.client_id}`
		);

		return { errors: [] };
	},
	/**
	 * Remove an URL reference.
	 */
	removeUrl: async ({ locals, url, params: { uuid }, ...event }) => {
		const { details, currentUser } = await getActionData(locals, uuid);

		const id = Number(url.searchParams.get('id'));
		if (isNaN(id)) {
			return fail(400, { errors: ['invalidUrlId'] });
		}

		await OAuth2Clients.deleteUrl(details, id);

		await Audit.insertRequest(
			AuditAction.OAUTH2_UPDATE,
			event,
			currentUser,
			`delete url\nclient_id=${details.client_id}`
		);

		return { errors: [] };
	},
	/**
	 * Add an URL reference.
	 */
	addUrl: async ({ locals, request, params: { uuid }, getClientAddress }) => {
		const { details, currentUser } = await getActionData(locals, uuid);

		const body = await request.formData();
		const { type, url } = Changesets.only(['type', 'url'], body);
		const urlType = type as OAuth2ClientURLType;
		if (!type || !OAuth2Clients.availableUrlTypes.includes(urlType)) {
			return fail(400, { errors: ['invalidUrlType'] });
		}

		if (!url) {
			return fail(400, { errors: ['invalidUrl'] });
		}

		// Do not allow adding multiple URLs of the same type, except for redirects.
		const existingURLs = (details as OAuth2ClientAdminListItem).urls;
		const existingOfType = existingURLs.filter((entry) => entry.type === type).length;
		const maxOfType =
			type === OAuth2ClientURLType.REDIRECT_URI ? OAUTH2_MAX_REDIRECTS : OAUTH2_MAX_URLS;
		if (maxOfType < existingOfType) {
			return fail(400, { errors: ['illegalUrl'] });
		}

		await OAuth2Clients.addUrl(details, urlType, url);

		await Audit.insertRequest(
			AuditAction.OAUTH2_UPDATE,
			{ request, getClientAddress },
			currentUser,
			`add url\nclient_id=${details.client_id}`
		);

		return { errors: [] };
	},
	/**
	 * Remove a privilege reference.
	 */
	removePrivilege: async ({ locals, url, params: { uuid }, ...event }) => {
		const { details, currentUser } = await getActionData(locals, uuid);

		const id = Number(url.searchParams.get('id'));
		if (isNaN(id)) {
			return fail(400, { errors: ['invalidPrivilegeId'] });
		}

		await OAuth2Clients.deletePrivilege(details, id);

		await Audit.insertRequest(
			AuditAction.OAUTH2_UPDATE,
			event,
			currentUser,
			`remove privilege\nclient_id=${details.client_id}`
		);

		return { errors: [] };
	},
	/**
	 * Add a privilege reference.
	 */
	addPrivilege: async ({ locals, request, params: { uuid }, getClientAddress }) => {
		const { details, currentUser } = await getActionData(locals, uuid);

		const body = await request.formData();
		const { name } = Changesets.only(['name'], body);

		if (!name || !privilegeRegex.test(name)) {
			return fail(400, { errors: ['invalidPrivilege'] });
		}

		await OAuth2Clients.addPrivilege(details, name);

		await Audit.insertRequest(
			AuditAction.OAUTH2_UPDATE,
			{ request, getClientAddress },
			currentUser,
			`add privilege\nclient_id=${details.client_id}`
		);

		return { errors: [] };
	},
	/**
	 * Update the OAuth2 Client allowed grants list.
	 */
	grants: async ({ locals, request, params: { uuid }, getClientAddress }) => {
		const { details, fullPrivileges, currentUser } = await getActionData(locals, uuid);

		const allowedGrants = fullPrivileges
			? OAuth2Clients.availableGrantTypes
			: OAuth2Clients.userSetGrants;

		const body = await request.formData();
		const values = Array.from(body.keys());
		values.unshift('authorization_code');
		const deduplicatedAllowedGrants = values.filter(
			(value, index, array) => allowedGrants.includes(value) && array.indexOf(value) === index
		);

		await OAuth2Clients.update(details, {
			grants: deduplicatedAllowedGrants.join(' ')
		});

		await Audit.insertRequest(
			AuditAction.OAUTH2_UPDATE,
			{ request, getClientAddress },
			currentUser,
			`grants\nclient_id=${details.client_id}`
		);

		return { errors: [] };
	},
	/**
	 * Update the OAuth2 Client allowed scopes list.
	 */
	scopes: async ({ locals, request, params: { uuid }, getClientAddress }) => {
		const { details, fullPrivileges, currentUser } = await getActionData(locals, uuid);

		const allowedScopes = fullPrivileges
			? OAuth2Clients.availableScopes
			: OAuth2Clients.userSetScopes;

		const body = await request.formData();
		const values = Array.from(body.keys());
		values.unshift('profile');
		const deduplicatedAllowedScopes = values.filter(
			(value, index, array) => allowedScopes.includes(value) && array.indexOf(value) === index
		);

		await OAuth2Clients.update(details, {
			scope: OAuth2Clients.joinScope(deduplicatedAllowedScopes)
		});

		await Audit.insertRequest(
			AuditAction.OAUTH2_UPDATE,
			{ request, getClientAddress },
			currentUser,
			`scopes\nclient_id=${details.client_id}`
		);

		return { errors: [] };
	},
	/**
	 * Update the OAuth2 Client icon picture.
	 */
	avatar: async ({ request, locals, params: { uuid }, getClientAddress }) => {
		const { currentUser, details } = await getActionData(locals, uuid);

		const formData = Object.fromEntries(await request.formData());
		if (!(formData.file as File)?.name || (formData.file as File).name === 'undefined') {
			return fail(400, {
				errors: ['noFile']
			});
		}

		const { file } = formData as { file: File };

		await Uploads.saveClientAvatar(details, currentUser, file);

		await Audit.insertRequest(
			AuditAction.OAUTH2_UPDATE,
			{ request, getClientAddress },
			currentUser,
			`upload avatar\nclient_id=${details.client_id}`
		);

		return { errors: [] };
	},
	/**
	 * Delete the OAuth2 Client icon picture.
	 */
	removeAvatar: async ({ locals, params: { uuid }, ...event }) => {
		const { details, currentUser } = await getActionData(locals, uuid);

		await Uploads.removeClientAvatar(details);

		await Audit.insertRequest(
			AuditAction.OAUTH2_UPDATE,
			event,
			currentUser,
			`remove avatar\nclient_id=${details.client_id}`
		);

		return { errors: [] };
	},
	/**
	 * Invite a new manager to the OAuth2 Client.
	 * Managers can do most changes the client's owner can.
	 *
	 * Only the owner or admin can do it.
	 */
	invite: async (event) => {
		const {
			locals,
			request,
			params: { uuid }
		} = event;
		const { currentUser, details, fullPrivileges } = await getActionData(locals, uuid);

		const body = await request.formData();
		const { email } = Changesets.only(['email'], body);

		if (!email || !emailRegex.test(email)) {
			return fail(400, { errors: ['invalidEmail'] });
		}

		if (details.ownerId !== currentUser.id && !fullPrivileges) {
			return fail(403, { errors: ['forbidden'] });
		}

		const managers = await OAuth2Clients.getManagers(details);
		if (managers.some((entry) => entry.email.toLowerCase() === email.toLowerCase())) {
			return fail(400, { errors: ['emailExists'] });
		}

		if (!fullPrivileges && (await inviteLimiter.isLimited(event))) {
			return fail(429, { errors: ['tooManyTimes'] });
		}

		await OAuth2Clients.sendManagerInvitationEmail(details as OAuth2Client, currentUser, email);

		await Audit.insertRequest(
			AuditAction.OAUTH2_INVITE,
			event,
			currentUser,
			`invite\nclient_id=${details.client_id}\nemail=${email}`
		);

		return { errors: [] };
	},
	/**
	 * Remove a manager form the OAuth2 Client.
	 *
	 * Only the owner or admin can do it.
	 */
	removeManager: async ({ locals, url, params: { uuid }, ...event }) => {
		const { details, currentUser, fullPrivileges } = await getActionData(locals, uuid);

		const id = Number(url.searchParams.get('id'));
		if (isNaN(id)) {
			return fail(400, { errors: ['invalidManagerId'] });
		}

		if (details.ownerId !== currentUser.id && !fullPrivileges) {
			return fail(403, { errors: ['forbidden'] });
		}

		await OAuth2Clients.removeManager(details, id);
		await Audit.insertRequest(
			AuditAction.OAUTH2_UPDATE,
			event,
			currentUser,
			`remove manager\nclient_id=${details.client_id}`
		);

		return { errors: [] };
	},
	/**
	 * Update client JWKs
	 */
	jwks: async ({ locals, request, params: { uuid }, getClientAddress }) => {
		const { currentUser, details } = await getActionData(locals, uuid);
		const body = await request.formData();
		const value = body.get('jwks') as string;
		if (!value) {
			return fail(403, { errors: ['jwksRequired'] });
		}

		try {
			await OAuth2Clients.updateJwks(details, value);
		} catch {
			return fail(403, { errors: ['invalidJwks'] });
		}

		await Audit.insertRequest(
			AuditAction.OAUTH2_UPDATE,
			{ request, getClientAddress },
			currentUser,
			`update jwks\nclient_id=${details.client_id}`
		);

		return { errors: [] };
	}
};

export const load = async ({ params: { uuid }, parent }) => {
	const { user } = await parent();
	const currentUser = await Users.getBySession(user);
	AdminUtils.checkPrivileges(user, ['admin:oauth2', 'self:oauth2']);

	const fullPrivileges = hasPrivileges(user.privileges, ['admin:oauth2']);

	const {
		list: [details]
	} = await OAuth2Clients.getClientByAdminUser(currentUser as User, {
		clientId: uuid,
		listAll: fullPrivileges,
		omitSecret: false
	});

	if (!details) {
		return error(404, 'Client not found');
	}

	const privileges = await Users.getAvailablePrivileges(details.id);
	const users = await OAuth2Clients.getAuthorizedUsers(details as OAuth2Client);
	const managers = await OAuth2Clients.getManagers(details as OAuth2Client);

	const availableGrants: string[] = [];
	if (fullPrivileges) {
		availableGrants.push(...OAuth2Clients.availableGrantTypes);
	} else {
		availableGrants.push(...OAuth2Clients.userSetGrants);

		// Implicit grants can be allowed additionally.
		if (hasPrivileges(user.privileges, ['self:oauth2:implicit'])) {
			availableGrants.push(...OAuth2Clients.implicitGrantTypes);
		}
	}

	const availableScopes = fullPrivileges
		? OAuth2Clients.availableScopes
		: OAuth2Clients.userSetScopes;

	return {
		users,
		managers,
		availableUrls: OAuth2Clients.availableUrlTypes,
		availablePrivileges: privileges,
		availableGrants,
		availableScopes,
		fullPrivileges,
		details
	};
};
